from .fixed import FixedQuota
from .quota import Quota
from .range import RangeQuota
from .registry import QUOTA_REGISTRY

QUOTA_REGISTRY.register("fixed", FixedQuota)
QUOTA_REGISTRY.register("const", FixedQuota)
QUOTA_REGISTRY.register("range", RangeQuota)
QUOTA_REGISTRY.register("random", RangeQuota)

__all__ = ["Quota", "FixedQuota", "RangeQuota", "QUOTA_REGISTRY"]



import itertools
from typing import Any, List

from prompt_collapse.state import State

from .constant import ConstContent
from .content import Content
from .registry import CONTENT_REGISTRY


class SequenceContent(Content):
    def __init__(
        self,
        content_entries: List[Content],
    ) -> None:
        self._content_entries = content_entries

    def apply(self, local_state: State) -> List[str]:
        return list(
            itertools.chain.from_iterable(
                [entry.apply(local_state) for entry in self._content_entries]
            )
        )

    @classmethod
    def from_spec(cls, spec: Any) -> "SequenceContent":
        assert isinstance(spec, dict), f"Invalid spec for SequenceContent: {spec}"

        values = spec["values"]

        return cls([cls._parse_content(content_spec) for content_spec in values])

    @staticmethod
    def _parse_content(spec: Any) -> Content:
        if isinstance(spec, (str, int, float)):
            return ConstContent(str(spec))

        CONTENT_REGISTRY.parse(spec)



import random
from typing import Any

from prompt_collapse.state import State

from .transformation import Transformation


class RandomIntTransformation(Transformation):
    def __init__(self, min_value: int, max_value: int) -> None:
        self._min_value = min_value
        self._max_value = max_value

    @classmethod
    def from_spec(cls, spec: Any) -> "RandomIntTransformation":
        min_value = spec.get("min")
        max_value = spec.get("max")

        return cls(min_value, max_value)

    def apply(self, value: Any, local_state: State, global_state: State) -> Any:
        """
        Returning a random integer between the min and max values.
        :param value: Ignored value
        :param local_state: Local state
        :param global_state: Global state
        :return: Random integer between min and max values
        """
        return random.randint(self._min_value, self._max_value)



from typing import Any

from prompt_collapse.state import State

from .declaration import Declaration


class IncrementDeclaration(Declaration):
    def __init__(self, name: str) -> None:
        self._name = name

    def name(self) -> str:
        return self._name

    def apply(self, local_state: State, global_state: State) -> Any:
        value = global_state.get(self._name)

        global_state.update(self._name, value + 1)

        return value + 1

    @classmethod
    def from_spec(cls, spec: Any) -> "IncrementDeclaration":
        assert isinstance(spec, dict), f"Invalid spec for IncrementDeclaration: {spec}"

        return cls(spec["variable"])



from typing import Any

from prompt_collapse.state import State

from ..distributions import WeightedDistribution
from .transformation import Transformation


class RandomValueTransformation(Transformation):
    def __init__(self, distribution: WeightedDistribution):
        self._distribution = distribution

    @classmethod
    def from_spec(cls, spec: Any) -> "RandomValueTransformation":
        values = spec.get("values")

        return cls(WeightedDistribution.from_spec(values))

    def apply(self, value: Any, local_state: State, global_state: State) -> Any:
        """
        Returning the value stored in the transformation, ignoring the provided value.

        :param value: The value to ignore
        :param local_state: The local state
        :param global_state: The global state
        :return: The value stored in the transformation
        """
        index = self._distribution.apply(k=1)[0]

        return self._distribution.get_items()[index]



from typing import Any

from prompt_collapse.state import State

from .transformation import Transformation


class IdentityTransformation(Transformation):
    @classmethod
    def from_spec(cls, spec: Any) -> "IdentityTransformation":
        if spec is not None:
            raise ValueError(
                f"Identity transformation expects no arguments, but {spec} was provided"
            )

        return cls()

    def apply(self, value: Any, local_state: State, global_state: State) -> Any:
        """
        Returning the value as is, without any modifications.

        :param value: The value to return
        :param local_state: The local state
        :param global_state: The global state
        :return: The original value
        """
        return value



from typing import Any

from prompt_collapse.state import State

from .requirement import Requirement


class EqualsRequirement(Requirement):
    def __init__(
        self,
        variable: str,
        value: Any,
    ) -> None:
        self._variable = variable
        self._value = value

    def apply(self, global_state: State) -> bool:
        value = global_state.get(self._variable)

        return value == self._value

    @classmethod
    def from_spec(cls, spec: Any) -> "EqualsRequirement":
        assert isinstance(spec, dict), "EqualsRequirement spec must be a dict"

        variable = spec.get("variable")
        value = spec.get("value")

        assert isinstance(
            variable, str
        ), "EqualsRequirement spec must have a string variable name"

        return cls(variable, value)



from abc import ABC, abstractmethod
from typing import Any, List, Type, TypeVar

T = TypeVar("T", bound="Distribution")


class Distribution(ABC):
    @classmethod
    @abstractmethod
    def from_spec(cls: Type[T], spec: Any) -> T:
        raise NotImplementedError

    @abstractmethod
    def apply(self, k: int = 1) -> List[int]:
        raise NotImplementedError

    @abstractmethod
    def get_items(self) -> dict[int, Any]:
        raise NotImplementedError

    @abstractmethod
    def get_item(self, index: int) -> Any:
        raise NotImplementedError



from typing import Any

from prompt_collapse.state import State

from .requirement import Requirement


class ExistsRequirement(Requirement):
    def __init__(
        self,
        variable: str,
    ) -> None:
        self._variable = variable

    def apply(self, global_state: State) -> bool:
        return global_state.exists(self._variable)

    @classmethod
    def from_spec(cls, spec: Any) -> "ExistsRequirement":
        assert isinstance(spec, dict), "GreaterThanRequirement spec must be a dict"

        variable = spec.get("variable")

        assert isinstance(
            variable, str
        ), "ExistsRequirement spec must have a string variable name"

        return cls(variable)



from typing import Any, Optional

from prompt_collapse.common.utils import apply_parameters, find_parameters
from prompt_collapse.state import State

from .declaration import Declaration


class EmitDeclaration(Declaration):
    def __init__(
        self,
        name: str,
        value: Any,
        cast_to: Optional[str] = None,
    ) -> None:
        self._name = name
        self._value = value
        self._cast_to = cast_to

        self._parameters = find_parameters(str(value))

    def name(self) -> str:
        return self._name

    def apply(self, local_state: State, global_state: State) -> Any:
        value = apply_parameters(self._value, self._parameters, local_state)

        if self._cast_to is not None:
            value = self._cast(value)

        global_state.set(self._name, value)

        return value

    def _cast(self, value: Any) -> Any:
        if self._cast_to == "int":
            return int(value)
        elif self._cast_to == "float":
            return float(value)
        elif self._cast_to == "str":
            return str(value)
        else:
            raise ValueError(f"Invalid cast_to value: {self._cast_to}")

    @classmethod
    def from_spec(cls, spec: Any) -> "EmitDeclaration":
        assert isinstance(spec, dict), f"Invalid spec for EmitDeclaration: {spec}"

        return cls(spec["variable"], spec.get("value", True), spec.get("cast_to"))



from abc import ABC, abstractmethod
from typing import Any, Type, TypeVar

T = TypeVar("T", bound="Quota")


class Quota(ABC):
    @abstractmethod
    def apply(self) -> int:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_spec(cls: Type[T], spec: Any) -> T:
        raise NotImplementedError



import itertools
from typing import Any, List

from prompt_collapse.common.distributions import WeightedDistribution
from prompt_collapse.common.quota import QUOTA_REGISTRY, FixedQuota, Quota
from prompt_collapse.state import State

from .constant import ConstContent
from .content import Content
from .registry import CONTENT_REGISTRY


class SomeOfContent(Content):
    def __init__(
        self,
        distribution: WeightedDistribution,
        index_to_content: dict[int, Content],
        quota: Quota,
    ) -> None:
        self._distribution = distribution
        self._index_to_content = index_to_content
        self._quota = quota

    def apply(self, local_state: State) -> List[str]:
        count = self._quota.apply()
        indices = self._distribution.apply(k=count)

        return list(
            itertools.chain.from_iterable(
                [self._index_to_content[index].apply(local_state) for index in indices]
            )
        )

    @classmethod
    def from_spec(cls, spec: Any) -> "SomeOfContent":
        assert isinstance(spec, (list, dict)), f"Invalid spec for SomeOfContent: {spec}"

        quota: Quota = FixedQuota(1)

        if isinstance(spec, dict):
            quota = cls._parse_quota(spec.get("quota"))
            spec = spec["values"]

        distribution = WeightedDistribution.from_spec(spec)

        index_to_content = {
            index: cls._parse_content(content_spec)
            for index, content_spec in distribution.get_items().items()
        }

        return cls(distribution, index_to_content, quota)

    @staticmethod
    def _parse_content(spec: Any) -> Content:
        if isinstance(spec, (str, int, float)):
            return ConstContent(str(spec))

        CONTENT_REGISTRY.get(spec["name"]).from_spec(spec)

        raise ValueError(f"Invalid spec for ConstContent: {spec}")

    @staticmethod
    def _parse_quota(spec: Any) -> Quota:
        if spec is None:
            return FixedQuota(1)

        if isinstance(spec, (int, str, float)):
            return FixedQuota(int(spec))

        return QUOTA_REGISTRY.parse(spec)



import random
from typing import List, Optional, Callable

from .state import State
from .component import Component
from .repository import ComponentRepository

ComponentFilter = Callable[[Component], bool]


class Solver:
    def __init__(
        self,
        repository: ComponentRepository,
    ) -> None:
        self._repository = repository

        self._unused_components: List[Component] = []
        self._context: List[Component] = []
        self._prompt_fragments: List[str] = []
        self._state = State()

    def generate_prompt(
        self, initial_tags: List[str]
    ) -> str:
        """
        Generates a prompt based on the initial tags

        :param initial_tags: Initial tags to generate the prompt
        :return: Generated prompt
        """
        self._reset()

        self._populate_with_initial_tags(initial_tags)

        while self._add_random_compatible_component():
            pass

        return ", ".join(set(self._prompt_fragments))

    def _populate_with_initial_tags(self, initial_tags: List[str]) -> None:
        """
        Populates the context with components that match the initial tags

        :param initial_tags: Initial tags to populate the context
        :return: None
        """
        initial_tags = list(set(initial_tags))
        random.shuffle(initial_tags)

        for tag in initial_tags:
            self._add_random_compatible_component(
                lambda c: tag in c.tags
            )

    def _add_random_compatible_component(self, extra_condition: Optional[ComponentFilter] = None) -> bool:
        """
        Adds a random compatible component to the context

        :return: None
        """
        components = self._get_compatible_components(extra_condition=extra_condition)

        if not components:
            return False

        component = random.choice(components)

        self._add_component(component)

        return True

    def _add_component(self, component: Component) -> None:
        """
        Adds a component to the context

        :param component: Component to add
        :return: None
        """
        self._context.append(component)
        self._unused_components.remove(component)

        local_state = component.build_local_state(self._state)

        values = component.apply(local_state, self._state)

        self._prompt_fragments.extend(values)

    def _get_compatible_components(self, extra_condition: Optional[ComponentFilter] = None) -> List[Component]:
        """
        Returns a list of compatible components

        :return: List of compatible components
        """
        components = [component for component in self._unused_components if component.check(self._state)]

        if extra_condition:
            components = [component for component in components if extra_condition(component)]

        return components

    def _reset(self) -> None:
        """
        Resets the context

        :return: None
        """
        self._context = []
        self._prompt_fragments = []
        self._state = State()
        self._unused_components = self._repository.get_all()

    @property
    def repository(self) -> ComponentRepository:
        return self._repository



from typing import Any

from prompt_collapse.common.transformations import (
    TRANSFORMATION_REGISTRY,
    Transformation,
)
from prompt_collapse.state import State


class Parameter:
    def __init__(self, name: str, transformation: Transformation) -> None:
        self._name = name
        self._transformation = transformation

    def name(self) -> str:
        return self._name

    def apply(self, local_state: State, global_state: State) -> Any:
        value = self._transformation.apply(None, local_state, global_state)
        local_state.set(self._name, value)

        return value

    @classmethod
    def from_spec(cls, spec: Any) -> "Parameter":
        if not isinstance(spec, dict):
            raise ValueError(f"Expected a dictionary, but got {spec}")

        name = spec["name"]
        transformation_spec = spec["transform"]

        transformation = TRANSFORMATION_REGISTRY.get(
            transformation_spec.get("name")
        ).from_spec(transformation_spec)

        return cls(name, transformation)

    def __repr__(self) -> str:
        return f"Parameter(name={self._name}, transformation={self._transformation})"

    def __str__(self) -> str:
        return repr(self)

    def __hash__(self) -> int:
        return hash(str(self))



from abc import ABC, abstractmethod
from typing import Any, Type, TypeVar

from prompt_collapse.state import State

T = TypeVar("T", bound="Declaration")


class Declaration(ABC):
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def apply(self, local_state: State, global_state: State) -> Any:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_spec(cls: Type[T], spec: Any) -> T:
        raise NotImplementedError



import random
from typing import Any

from .quota import Quota


class RangeQuota(Quota):
    def __init__(self, min_count: int, max_count: int) -> None:
        self.min_count = min_count
        self.max_count = max_count

    def apply(self) -> int:
        return random.randint(self.min_count, self.max_count)

    @classmethod
    def from_spec(cls, spec: Any) -> "RangeQuota":
        min_val = int(spec.get("min", 0))
        max_val = int(spec.get("max", 0))

        return cls(min_val, max_val)

    def __repr__(self) -> str:
        return "RangeQuota(min=%s, max=%s)" % (self.min_count, self.max_count)

    def __str__(self) -> str:
        return repr(self)

    def __hash__(self) -> int:
        return hash(str(self))



from .all_of import AllOfRequirement
from .equals import EqualsRequirement
from .exists import ExistsRequirement
from .greater_than import GreaterThanRequirement
from .less_than import LessThanRequirement
from .none_of import NoneOfRequirement
from .registry import REQUIREMENT_REGISTRY
from .requirement import Requirement
from .some_of import SomeOfRequirement

REQUIREMENT_REGISTRY.register("less_than", LessThanRequirement)
REQUIREMENT_REGISTRY.register("greater_than", GreaterThanRequirement)
REQUIREMENT_REGISTRY.register("exists", ExistsRequirement)
REQUIREMENT_REGISTRY.register("all_of", AllOfRequirement)
REQUIREMENT_REGISTRY.register("some_of", SomeOfRequirement)
REQUIREMENT_REGISTRY.register("none_of", NoneOfRequirement)
REQUIREMENT_REGISTRY.register("equals", EqualsRequirement)

REQUIREMENT_REGISTRY.register("lt", LessThanRequirement)
REQUIREMENT_REGISTRY.register("gt", GreaterThanRequirement)
REQUIREMENT_REGISTRY.register("and", AllOfRequirement)
REQUIREMENT_REGISTRY.register("or", SomeOfRequirement)
REQUIREMENT_REGISTRY.register("not", NoneOfRequirement)

__all__ = [
    "Requirement",
    "LessThanRequirement",
    "GreaterThanRequirement",
    "ExistsRequirement",
    "AllOfRequirement",
    "SomeOfRequirement",
    "NoneOfRequirement",
    "EqualsRequirement",
    "REQUIREMENT_REGISTRY",
]



from typing import Any

from prompt_collapse.state import State

from .requirement import Requirement


class GreaterThanRequirement(Requirement):
    def __init__(
        self,
        variable: str,
        value: int,
    ) -> None:
        self._variable = variable
        self._value = value

    def apply(self, global_state: State) -> bool:
        if not global_state.exists(self._variable):
            return False

        value = global_state.get(self._variable)

        return value > self._value

    @classmethod
    def from_spec(cls, spec: Any) -> "GreaterThanRequirement":
        assert isinstance(spec, dict), "GreaterThanRequirement spec must be a dict"

        variable = spec.get("variable")
        value = spec.get("value")

        assert isinstance(variable, str) and isinstance(
            value, int
        ), "GreaterThanRequirement spec must have a string variable and an integer value"

        return cls(variable, value)



from typing import Any, List, Set

from prompt_collapse.content import CONTENT_REGISTRY, Content, SequenceContent, ConstContent
from prompt_collapse.declaration import DECLARATION_REGISTRY, Declaration
from prompt_collapse.parameters import Parameter
from prompt_collapse.requirement import (
    REQUIREMENT_REGISTRY,
    AllOfRequirement,
    Requirement,
)
from prompt_collapse.state import State


class Component:
    def __init__(
        self,
        alias: str,
        parameters: List[Parameter],
        declares: List[Declaration],
        requires: List[Requirement],
        generates: List[Content],
        tags: List[str],
    ) -> None:
        self._alias = alias
        self._parameters = parameters
        self._declarations = declares
        self._requirement = AllOfRequirement(requires)
        self._content = SequenceContent(generates)
        self._tags = set(tags)

    @property
    def alias(self) -> str:
        return self._alias

    @property
    def tags(self) -> Set[str]:
        return self._tags.copy()

    def build_local_state(self, global_state: State) -> State:
        local_state = State()

        for parameter in self._parameters:
            parameter.apply(local_state, global_state)

        return local_state

    def apply(self, local_state: State, global_state: State) -> list[str]:
        for declaration in self._declarations:
            declaration.apply(local_state, global_state)

        return self._content.apply(local_state)

    def check(self, global_state: State) -> bool:
        return self._requirement.apply(global_state)

    @classmethod
    def from_spec(cls, spec: dict[str, Any]) -> "Component":
        alias = cls._get_alias(spec)
        parameters = cls._get_parameters(spec)
        declares = cls._get_declarations(spec)
        requires = cls._get_requirements(spec)
        generates = cls._get_content(spec)
        tags = cls._get_tags(spec)

        return cls(
            alias,
            parameters,
            declares,
            requires,
            generates,
            tags,
        )

    @staticmethod
    def _get_alias(spec: dict[str, Any]) -> str:
        alias = spec.get("alias")
        assert isinstance(alias, str), "Component spec must have a string alias"
        return alias

    @staticmethod
    def _get_parameters(spec: dict[str, Any]) -> List[Parameter]:
        parameters = spec.get("parameters", [])
        assert isinstance(parameters, list), "Component spec parameters must be a list"

        return [Parameter.from_spec(p) for p in parameters]

    @staticmethod
    def _get_declarations(spec: dict[str, Any]) -> List[Declaration]:
        declares = spec.get("declares", [])
        assert isinstance(declares, list), "Component spec declares must be a list"

        return [DECLARATION_REGISTRY.parse(d) for d in declares]

    @staticmethod
    def _get_requirements(spec: dict[str, Any]) -> List[Requirement]:
        requires = spec.get("requires", [])
        assert isinstance(requires, list), "Component spec requires must be a list"

        return [REQUIREMENT_REGISTRY.parse(r) for r in requires]

    @staticmethod
    def _get_content(spec: dict[str, Any]) -> List[Content]:
        generates = spec.get("generates", [])
        assert isinstance(generates, list), "Component spec generates must be a list"

        content = []

        for entry in generates:
            if not isinstance(entry, dict):
                content.append(ConstContent(entry))

            else:
                content.append(CONTENT_REGISTRY.parse(entry))

        return content

    @staticmethod
    def _get_tags(spec: dict[str, Any]) -> List[str]:
        tags = spec.get("tags", [])
        assert isinstance(tags, list), "Component spec tags must be a list"

        return tags

    def __repr__(self) -> str:
        return (
            f"Component("
            f"alias={self._alias}"
            f", parameters={self._parameters}"
            f", declares={self._declarations}"
            f", requires={self._requirement}"
            f", generates={self._content}"
            f", tags={self._tags})"
        )

    def __str__(self) -> str:
        return repr(self)

    def __hash__(self) -> int:
        return hash(str(self))



from .declaration import Declaration
from .decrement import DecrementDeclaration
from .emit import EmitDeclaration
from .increment import IncrementDeclaration
from .overwrite import OverwriteDeclaration
from .registry import DECLARATION_REGISTRY

DECLARATION_REGISTRY.register("emit", EmitDeclaration)
DECLARATION_REGISTRY.register("decrement", DecrementDeclaration)
DECLARATION_REGISTRY.register("increment", IncrementDeclaration)
DECLARATION_REGISTRY.register("overwrite", OverwriteDeclaration)

DECLARATION_REGISTRY.register("set", EmitDeclaration)
DECLARATION_REGISTRY.register("dec", DecrementDeclaration)
DECLARATION_REGISTRY.register("inc", IncrementDeclaration)
DECLARATION_REGISTRY.register("update", OverwriteDeclaration)

__all__ = [
    "Declaration",
    "EmitDeclaration",
    "DecrementDeclaration",
    "IncrementDeclaration",
    "OverwriteDeclaration",
    "DECLARATION_REGISTRY",
]



from typing import Any

from prompt_collapse.state import State

from .transformation import Transformation


class LookupTransformation(Transformation):
    def __init__(self, variable: str) -> None:
        self._variable = variable

    @classmethod
    def from_spec(cls, spec: Any) -> "LookupTransformation":
        if not isinstance(spec, dict):
            raise ValueError(f"Expected a dictionary, but got {spec}")

        variable = spec.get("variable")

        if not isinstance(variable, str):
            raise ValueError(f"Expected a string variable, but got {variable}")

        return cls(variable)

    def apply(self, value: Any, local_state: State, global_state: State) -> Any:
        """
        Replacing the current value with the value of the specified variable in one of the provided states.
        The lookup order is local state -> global state, similar to the variable lookup order in Python.

        :param value: The value to replace
        :param local_state: The local state
        :param global_state: The global state
        :return: The value of the specified variable
        """

        if local_state.exists(self._variable):
            return local_state.get(self._variable)

        if global_state.exists(self._variable):
            return global_state.get(self._variable)

        raise ValueError(
            f"Variable {self._variable} not found in local or global state"
        )



from .constant import ConstContent
from .content import Content
from .registry import CONTENT_REGISTRY
from .sequence import SequenceContent
from .some_of import SomeOfContent

CONTENT_REGISTRY.register("const", ConstContent)
CONTENT_REGISTRY.register("constant", ConstContent)
CONTENT_REGISTRY.register("some_of", SomeOfContent)
CONTENT_REGISTRY.register("choice", SomeOfContent)
CONTENT_REGISTRY.register("sequence", SequenceContent)
CONTENT_REGISTRY.register("seq", SequenceContent)

__all__ = [
    "Content",
    "ConstContent",
    "SomeOfContent",
    "SequenceContent",
    "CONTENT_REGISTRY",
]



from abc import ABC, abstractmethod
from typing import Any, List, Type, TypeVar

from prompt_collapse.state import State

T = TypeVar("T", bound="Content")


class Content(ABC):
    @abstractmethod
    def apply(self, local_state: State) -> List[str]:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_spec(cls: Type[T], spec: Any) -> T:
        raise NotImplementedError



from typing import Any

from prompt_collapse.common.utils import find_parameters
from prompt_collapse.state import State

from .declaration import Declaration


class OverwriteDeclaration(Declaration):
    def __init__(
        self,
        name: str,
        value: Any,
    ) -> None:
        self._name = name
        self._value = value

        self._parameters = find_parameters(str(value))

    def name(self) -> str:
        return self._name

    def apply(self, local_state: State, global_state: State) -> Any:
        global_state.update(self._name, self._value)

        return self._value

    @classmethod
    def from_spec(cls, spec: Any) -> "OverwriteDeclaration":
        assert isinstance(spec, dict), f"Invalid spec for OverwriteDeclaration: {spec}"

        return cls(spec["variable"], spec["value"])



from abc import ABC, abstractmethod
from typing import Any, TypeVar

from prompt_collapse.state import State

T = TypeVar("T", bound="Transformation")


class Transformation(ABC):
    @classmethod
    @abstractmethod
    def from_spec(cls, spec: Any) -> "T":
        raise NotImplementedError

    @abstractmethod
    def apply(self, value: Any, local_state: State, global_state: State) -> Any:
        """
        Applying the transformation to the provided value (if any) given both the local and global state.
        Local state only exists during component evaluation, while global state is used throughout the entire
        prompt generation process.

        :param value: The value to apply the transformation to
        :param local_state: The local state obtained after "collapsing" component parameters
        :param global_state: The global state/context
        :return: The transformed value
        """
        raise NotImplementedError



from typing import Any, List

from prompt_collapse.state import State

from .registry import TRANSFORMATION_REGISTRY
from .transformation import Transformation


class SequenceTransformation(Transformation):
    def __init__(self, transformations: List[Transformation]) -> None:
        assert len(transformations) > 0, "Expected at least one transformation"

        self._transformations = transformations

    @classmethod
    def from_spec(cls, spec: Any) -> "SequenceTransformation":
        if not isinstance(spec, list):
            raise ValueError(f"Expected a list of transformations, but got {spec}")

        return cls(
            [
                cls._parse_transformation(transformation_spec)
                for transformation_spec in spec
            ]
        )

    def apply(self, value: Any, local_state: State, global_state: State) -> Any:
        """
        Applying each transformation in the sequence to the provided value, with the output of the previous
        transformation being the input of the next one.

        :param value: The initial value of the sequence
        :param local_state: The local state
        :param global_state: The global state
        :return: The final value after applying all transformations
        """
        for transformation in self._transformations:
            value = transformation.apply(value, local_state, global_state)

        return value

    @staticmethod
    def _parse_transformation(transformation_spec: Any) -> Transformation:
        return TRANSFORMATION_REGISTRY.parse(transformation_spec)



from typing import Any

from prompt_collapse.state import State

from .transformation import Transformation


class ConstantTransformation(Transformation):
    def __init__(self, value: Any):
        self._value = value

    @classmethod
    def from_spec(cls, spec: Any) -> "ConstantTransformation":
        return cls(spec)

    def apply(self, value: Any, local_state: State, global_state: State) -> Any:
        """
        Returning the value stored in the transformation, ignoring the provided value.

        :param value: The value to ignore
        :param local_state: The local state
        :param global_state: The global state
        :return: The value stored in the transformation
        """
        return self._value



import re
from typing import Any, List

from prompt_collapse.state import State


def find_parameters(value: str) -> List[str]:
    """
    Find all parameters in a string.
    Parameters are defined as words surrounded by curly braces.
    Essentially string placeholders.
    :param value: A string to search for parameters.
    :return: A list of parameters found in the string.
    """
    return [match[1:-1] for match in re.findall(r"{\w+}", value)]


def apply_parameters(value: Any, parameters: List[str], local_state: State) -> Any:
    """
    Apply parameters to a value.
    If the value is not a string, it is returned as is.

    Parameters are replaced in the value using the local state.
    :param value: A value to apply parameters to.
    :param parameters: A list of parameters to apply.
    :param local_state: The local state to use for parameter values.
    :return: The value with parameters applied.
    """
    if not isinstance(value, str):
        return value

    variables = {parameter: local_state.get(parameter) for parameter in parameters}

    return value.format(**variables)



from typing import Any

from .quota import Quota


class FixedQuota(Quota):
    def __init__(self, count: int) -> None:
        self.count = count

    def apply(self) -> int:
        return self.count

    @classmethod
    def from_spec(cls, spec: Any) -> "FixedQuota":
        return cls(int(spec))

    def __repr__(self) -> str:
        return "FixedQuota(count=%s)" % self.count

    def __str__(self) -> str:
        return repr(self)

    def __hash__(self) -> int:
        return hash(str(self))



import random
from typing import Any, List

from .distribution import Distribution


class WeightedDistribution(Distribution):
    def __init__(self, weights: List[tuple[float, Any]]) -> None:
        assert len(weights) > 0, "Cannot create an empty WeightedDistribution"

        weight_values, items = zip(*weights)
        indices = list(range(len(weight_values)))

        self._indices = indices
        self._weights = weight_values
        self._indices_to_items = dict(zip(indices, items))

    def apply(self, k: int = 1) -> List[int]:
        item_indices = random.choices(self._indices, weights=self._weights, k=k)

        return list(item_indices)[:k]

    def get_items(self) -> dict[int, Any]:
        return self._indices_to_items.copy()

    def get_item(self, index: int) -> Any:
        return self._indices_to_items[index]

    @classmethod
    def from_spec(cls, spec: Any) -> "WeightedDistribution":
        if not isinstance(spec, list):
            raise ValueError(f"Invalid spec for WeightedDistribution: {spec}")

        weights = [cls._load_weighted_item(item_spec) for item_spec in spec]

        return cls(weights)

    @staticmethod
    def _load_weighted_item(spec: Any) -> tuple[float, Any]:
        if not isinstance(spec, dict):
            raise ValueError(f"Invalid spec for WeightedDistribution: {spec}")

        weight = spec.get("weight", 1.0)
        value = spec.get("value")

        return float(weight), value



from typing import Any, List

from prompt_collapse.state import State

from .registry import REQUIREMENT_REGISTRY
from .requirement import Requirement


class AllOfRequirement(Requirement):
    def __init__(
        self,
        conditions: List[Requirement],
    ):
        self._conditions = conditions

    def apply(self, global_state: State) -> bool:
        return all(
            condition.apply(global_state) for condition in self._conditions
        )

    @classmethod
    def from_spec(cls, spec: Any) -> "AllOfRequirement":
        assert isinstance(spec, dict), "AllOfRequirement spec must be a dict"

        conditions = spec.get("conditions")

        return cls(
            [
                REQUIREMENT_REGISTRY.parse(condition_spec)
                for condition_spec in conditions
            ]
        )



from typing import Any, Optional, Tuple

from prompt_collapse.state import State

from .transformation import Transformation


class MatchTransformation(Transformation):
    def __init__(
        self,
        match_cases: dict[Optional[Any], Any],
        default: Optional[Any],
    ) -> None:
        assert len(match_cases) > 0, "Expected at least one match case"

        self._match_cases = match_cases
        self._default = default

    @classmethod
    def from_spec(cls, spec: Any) -> "MatchTransformation":
        if not isinstance(spec, dict):
            raise ValueError(f"Expected a list of transformations, but got {spec}")

        return cls(
            dict(
                cls._parse_transformation(transformation_spec)
                for transformation_spec in spec.get("cases", [])
            ),
            spec.get("default"),
        )

    def apply(self, value: Any, local_state: State, global_state: State) -> Any:
        """
        Matching the provided value to one of the cases and returning the corresponding value.
        If no match is found, the default value is returned.

        :param value: The value to match
        :param local_state: The local state
        :param global_state: The global state
        :return: The matched value or the default value
        """
        return self._match_cases.get(value, self._default)

    @staticmethod
    def _parse_transformation(transformation_spec: Any) -> Tuple[Optional[Any], Any]:
        if not isinstance(transformation_spec, dict):
            raise ValueError(f"Expected a dictionary, but got {transformation_spec}")

        match_value = transformation_spec.get("match")
        new_value = transformation_spec.get("value")

        if new_value is None:
            raise ValueError(f"Expected a value for the match case {match_value}")

        return match_value, new_value



from typing import Any, Generic, Type, TypeVar

T = TypeVar("T")


class TypeRegistry(Generic[T]):
    """
    A registry for transformations, allowing to register and retrieve transformations by alias.
    Actively used to resolve transformations from their aliases in the configuration during parsing.
    """

    def __init__(self):
        self._types: dict[str, Type[T]] = {}

    def register(self, alias: str, cls: Type[T]) -> None:
        self._types[alias] = cls

    def get(self, alias: str) -> Type[T]:
        if alias not in self._types:
            raise ValueError(f"Alias {alias} not found in registry")

        return self._types[alias]

    def parse(self, spec: Any) -> T:
        if not isinstance(spec, dict):
            raise ValueError(f"Expected a dictionary, but got {spec}")

        name = spec.get("name")

        if name is None:
            raise ValueError("Expected a name for the type")

        cls = self.get(name)

        return cls.from_spec(spec)



from typing import Any, List

from prompt_collapse.common.utils import apply_parameters, find_parameters
from prompt_collapse.state import State

from .content import Content


class ConstContent(Content):
    def __init__(self, value: str) -> None:
        self._value = value
        self._parameters = find_parameters(value)

    def apply(self, local_state: State) -> List[str]:
        return [apply_parameters(self._value, self._parameters, local_state)]

    @classmethod
    def from_spec(cls, spec: Any) -> "ConstContent":
        if isinstance(spec, (str, int, float)):
            return cls(str(spec))

        raise ValueError(f"Invalid spec for ConstContent: {spec}")



from typing import Any


class State:
    def __init__(self) -> None:
        self._variables: dict[str, Any] = {}

    def exists(self, variable: str) -> bool:
        return variable in self._variables

    def get(self, variable: str) -> Any:
        return self._variables.get(variable)

    def set(self, variable: str, value: Any) -> None:
        if variable in self._variables:
            if self._variables[variable] != value:
                raise ValueError(
                    f"Variable {variable} already exists with a different value"
                )

        self._variables[variable] = value

    def update(self, variable: str, value: Any) -> None:
        if variable not in self._variables:
            raise ValueError(f"Variable {variable} does not exist")

        self._variables[variable] = value



from .constant import ConstantTransformation
from .identity import IdentityTransformation
from .lookup import LookupTransformation
from .match import MatchTransformation
from .random_value import RandomValueTransformation
from .random_int import RandomIntTransformation
from .registry import TRANSFORMATION_REGISTRY
from .sequence import SequenceTransformation
from .transformation import Transformation

TRANSFORMATION_REGISTRY.register("lookup", LookupTransformation)
TRANSFORMATION_REGISTRY.register("constant", ConstantTransformation)
TRANSFORMATION_REGISTRY.register("identity", IdentityTransformation)
TRANSFORMATION_REGISTRY.register("sequence", SequenceTransformation)
TRANSFORMATION_REGISTRY.register("match", MatchTransformation)
TRANSFORMATION_REGISTRY.register("random_value", RandomValueTransformation)
TRANSFORMATION_REGISTRY.register("random_int", RandomIntTransformation)

TRANSFORMATION_REGISTRY.register("const", ConstantTransformation)
TRANSFORMATION_REGISTRY.register("seq", SequenceTransformation)
TRANSFORMATION_REGISTRY.register("id", IdentityTransformation)
TRANSFORMATION_REGISTRY.register("one_of", RandomValueTransformation)
TRANSFORMATION_REGISTRY.register("range", RandomIntTransformation)


__all__ = [
    "Transformation",
    "LookupTransformation",
    "ConstantTransformation",
    "IdentityTransformation",
    "SequenceTransformation",
    "MatchTransformation",
    "TRANSFORMATION_REGISTRY",
]



import json
import yaml
import os

from typing import List

from .component import Component


class ComponentRepository:
    def __init__(self):
        self._components: dict[str, Component] = {}

    def load(self, path: str) -> None:
        """
        Loads all component files recursively from a given path

        :param path: Path to a component collection
        :return: None
        """
        # List all files in the directory
        # For each file, parse all the components within it and add them to the repository
        filenames = self._list_files(path)

        for filename in filenames:
            components = self._parse_components(filename)

            for component in components:
                if component.alias in self._components:
                    raise ValueError(f"Component with alias {component.alias} already exists")

                self._components[component.alias] = component

    def get(self, alias: str) -> Component:
        if alias not in self._components:
            raise ValueError(f"Component with alias {alias} not found")

        return self._components[alias]

    def get_all(self) -> List[Component]:
        return list(self._components.values())

    def _parse_components(self, file_path: str) -> List[Component]:
        expected_extensions = [".yaml", ".yml", ".json"]

        if not any(file_path.endswith(ext) for ext in expected_extensions):
            return []

        if file_path.endswith(".json"):
            return self._parse_json(file_path)

        return self._parse_yaml(file_path)

    @staticmethod
    def _parse_json(file_path: str) -> List[Component]:
        with open(file_path, "r") as file:
            data = json.load(file)

        return [Component.from_spec(spec) for spec in data]

    @staticmethod
    def _parse_yaml(file_path: str) -> List[Component]:
        with open(file_path, "r") as file:
            data = yaml.safe_load(file)

        return [Component.from_spec(spec) for spec in data]

    @staticmethod
    def _list_files(path: str) -> List[str]:
        filenames = []

        for root, _, files in os.walk(path):
            for file in files:
                filenames.append(os.path.join(root, file))

        return filenames
